//+++ Overwritten by the version from nBMSL.

#include "STM.h"

SymTabMgr::SymTabMgr(){
  top_ = 0;
  scopeStack_[top_] = new GlobalEntry(string("GLOBAL")) ;
  isNameSpace_[top_] = false;
}

SymTabMgr::~SymTabMgr() { 
  // To be filled in later.
}

const SymTabEntry* 
SymTabMgr::lookUp(string name) const {
  const SymTabEntry *ste = 0;
  for (int i = top_; i >= 0; i--) {
    const SymTab* st = scopeStack_[i]->symTab();
    if ((st != nullptr) && ((ste = st->lookUp(name)) != nullptr))
      break;
  }
  return ste;
}


const SymTabEntry*
SymTabMgr::lookUpWithInheritance(string name) const {
  const SymTabEntry* ste = currentScope();
  // @@@@ Is this correct? Not checked yet.
  if ((ste != nullptr) && (ste->kind() == STRUCT_KIND)) {
    const StructEntry* se = (StructEntry*)ste;
    //list<StructEntry*> lse;
    //for (int i=se->nParents()-1;i >=0; i--)
    //lse.push_front((StructEntry*)se->parent(i));
    return se->lookUpWithInheritance(name,se);
  }
  return nullptr;
}

const SymTabEntry* 
SymTabMgr::lookUpInScope(string name, SymKind k) const {
  const SymTabEntry *ste = currentScope(k);
  if ((ste != nullptr) && (ste->symTab() != nullptr)) {
    ste = ste->symTab()->lookUp(name);
    return ste;
  }
  else return nullptr;
}


const SymTabEntry* 
SymTabMgr::currentScope() const {
  for (int i = top_; i >= 0; i--) {
    if (!isNameSpace_[i]) 
      return scopeStack_[i];
  }
  return nullptr;
}	

const SymTabEntry* 
SymTabMgr::currentScope(SymKind k) const {
  for (int i = top_; i >= 0; i--) {
    if (scopeStack_[i]->kind() == k) 
      return scopeStack_[i];
  }
  return nullptr;
}	


ErrorST 
SymTabMgr::enterScope(SymTabEntry* ste){
  if (ste == nullptr)
    return INVALID_SCOPE;
  if (++top_ >= MAX_SCOPE_STACK_SIZE) {
    errMsg("Too many nested levels of scope, exiting");
    exit(-1);
  }
  scopeStack_[top_] = ste;
  isNameSpace_[top_] = false;
  return OK;
}

ErrorST 
SymTabMgr::enterNameSpace(SymTabEntry* ste){
  ErrorST rv = enterScope(ste);
  isNameSpace_[top_] = true;
  return rv;
}

ErrorST 
SymTabMgr::leaveScope() {
  //cerr << "leaveScope called, top_=" <<  top_ << endl;
  if (top_ > 0) {
    top_--;
    return OK;
  }
  return INVALID_SCOPE;
}

ErrorST 
SymTabMgr::leaveScope(SymKind kind) {
  ErrorST rv = leaveToScope(kind);
  if (rv == OK)  
	return leaveScope();
  else return rv;
    }

ErrorST 
SymTabMgr::leaveToScope(SymKind kind) {
  ErrorST rv=OK;
  for (int i = top_; (i > 0) && (scopeStack_[i]->kind() != kind); i--)
	rv = leaveScope();
  return rv;
}

ErrorST 
SymTabMgr::insert(SymTabEntry* se) { 
  for (int i=top_; i >= 0; i--) 
    if (scopeStack_[i] != nullptr && !isNameSpace_[i])
      return (scopeStack_[i]->insert(se));
  
  return INVALID_SCOPE;
}
